/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include <png.h>

#include "gp2x_psp.h"
#include "psp_sdl.h"
#include "psp_global.h"
#include "psp_kbd.h"
#include "psp_go.h"
#include "gnugo.h"
#include "psp_menu.h"
#include "psp_play.h"

  static int psp_go_board_dirty = 1;
  static int psp_go_move_mode   = 1;
  static int psp_go_status_mode = 0;


void
psp_go_set_move_mode(int mode)
{
  psp_go_move_mode = mode;
}

static inline void
psp_go_get_pix_xy(int x, int y, int *px_pix, int *py_pix)
{
  int square_width_pix = GP2X_SDL_SCREEN_HEIGHT / (GGO.board_width + 1);
  int board_width_pix  = square_width_pix * (GGO.board_width - 1);
  int x_orig = (320 - board_width_pix) / 2;
  int y_orig = (GP2X_SDL_SCREEN_HEIGHT - board_width_pix) / 2;

  *px_pix = x_orig + (x * square_width_pix);
  *py_pix = y_orig + ((GGO.board_width - 1 - y) * square_width_pix);
}

static inline void
psp_go_display_handicap_xy(int x, int y)
{
  int x_pix = 0;
  int y_pix = 0;

  psp_go_get_pix_xy(x, y, &x_pix, &y_pix);

  psp_sdl_put_pixel(x_pix - 1, y_pix - 1, GGO.board_color);
  psp_sdl_put_pixel(x_pix + 1, y_pix - 1, GGO.board_color);
  psp_sdl_put_pixel(x_pix + 1, y_pix + 1, GGO.board_color);
  psp_sdl_put_pixel(x_pix - 1, y_pix + 1, GGO.board_color);
}

static inline void
psp_go_display_handicap()
{
  int boardsize = GGO.board_width;

  /* Display handicap */
  if (boardsize == 5) {

    psp_go_display_handicap_xy(1, 1);
    psp_go_display_handicap_xy(3, 1);
    psp_go_display_handicap_xy(1, 3);
    psp_go_display_handicap_xy(3, 3);
    psp_go_display_handicap_xy(2, 2);
    
    return;
  }

  if (!(boardsize%2)) {
    /* If the board size is even, no center handicap spots. */
    if (boardsize > 2 && boardsize < 12) {
      /* Place the outer 4 only. */
      psp_go_display_handicap_xy(2,2);
      psp_go_display_handicap_xy(boardsize - 3,2);
      psp_go_display_handicap_xy(2,boardsize - 3);
      psp_go_display_handicap_xy(boardsize-3,boardsize - 3);
    }
    else {
      /* Place the outer 4 only. */
      psp_go_display_handicap_xy(3,3);
      psp_go_display_handicap_xy(boardsize - 4,3);
      psp_go_display_handicap_xy(3, boardsize - 4);
      psp_go_display_handicap_xy(boardsize - 4, boardsize - 4);
    }
  }
  else {
    /* Uneven board size */
    if (boardsize > 2 && boardsize < 12) {
      /* Place the outer 4... */
      psp_go_display_handicap_xy(2,2);
      psp_go_display_handicap_xy(boardsize-3,2);
      psp_go_display_handicap_xy(2, boardsize-3);
      psp_go_display_handicap_xy(boardsize-3, boardsize-3);

      /* ...and the middle one. */
      psp_go_display_handicap_xy(boardsize/2, boardsize/2);
    }
    else {
      /* Place the outer 4... */
      psp_go_display_handicap_xy(3,3);
      psp_go_display_handicap_xy(boardsize - 4,3);
      psp_go_display_handicap_xy(3, boardsize - 4);
      psp_go_display_handicap_xy(boardsize - 4, boardsize - 4);

      /* ...and the inner 4... */
      psp_go_display_handicap_xy(3, boardsize/2);
      psp_go_display_handicap_xy(boardsize/2, 3);
      psp_go_display_handicap_xy(boardsize/2, boardsize-4);
      psp_go_display_handicap_xy(boardsize-4, boardsize/2);

      /* ...and the middle one. */
      psp_go_display_handicap_xy(boardsize/2, boardsize/2);
    }
  }
}

int
psp_go_is_computer_move()
{
  if ((! psp_go_move_mode) || GGO.game_over) return 0;

  if (GGO.go_game_mode == GGO_HUMAN_VS_COMPUTER) {
    return (GGO.gameinfo.computer_player == GGO.gameinfo.to_move);
  } else 
  if (GGO.go_game_mode == GGO_COMPUTER_VS_COMPUTER) {
    return 1;
  }
  return 0;
}

void
psp_go_display_board()
{
  char buffer[32];
  int  x_orig;
  int  y_orig;
  int  x_dst;
  int  y_dst;
  int  x_pix;
  int  y_pix;
  int  x;
  int  y;
  int  board_size_id;
  char letter;

  int square_width_pix = GP2X_SDL_SCREEN_HEIGHT / (GGO.board_width + 1);
  int board_width_pix  = square_width_pix * (GGO.board_width - 1);

  x_orig = (320 - board_width_pix) / 2;
  y_orig = (GP2X_SDL_SCREEN_HEIGHT - board_width_pix) / 2;

  if (GGO.board_width <=  8) board_size_id = 2;
  else 
  if (GGO.board_width <= 12) board_size_id = 1;
  else                       board_size_id = 0;

  psp_sdl_blit_board();

  psp_sdl_draw_rectangle(x_orig, y_orig,
                         board_width_pix, board_width_pix, GGO.board_color, 0);

  for (y = GGO.board_width; y > 0; y--) {
    x_dst = x_orig;
    y_dst = y_orig + ((GGO.board_width - y) * square_width_pix);

    if (y != 0)  {
      psp_sdl_fill_rectangle(x_dst, y_dst, board_width_pix, 0, GGO.board_color, 0);
    }

    sprintf(buffer, "%-2d", y);
    x_dst -= square_width_pix;
    if (GGO.board_width >= 10) x_dst -= square_width_pix;
    psp_sdl_print(x_dst, y_dst - 4, buffer, GGO.board_color);

    sprintf(buffer, "%2d", y);
    x_dst = x_orig + board_width_pix;
    x_dst += square_width_pix;
    psp_sdl_print(x_dst, y_dst - 4, buffer, GGO.board_color);
  }

  for (x = 0; x < GGO.board_width; x++) {
    x_dst = x_orig + (x * square_width_pix);
    y_dst = y_orig;

    letter = 'A' + x;
    if (letter >= 'I') letter++;

    psp_sdl_fill_rectangle(x_dst, y_dst, 0, board_width_pix, GGO.board_color, 0);

    y_dst = y_orig - (square_width_pix / 2) - 8;
    if (y_dst <= 0) y_dst = 1;
    psp_sdl_put_char(x_dst - 3, y_dst, GGO.board_color, 0, letter, 1, 0);

    y_dst = y_orig + board_width_pix;
    if (((y_dst + (square_width_pix / 2) + 10)) < GP2X_SDL_SCREEN_HEIGHT) {
      psp_sdl_put_char(x_dst - 3, y_dst + (square_width_pix / 2), GGO.board_color, 0, letter, 1, 0);
    }
  }

  psp_go_display_handicap();

  /* Display stones */
  for (y = 0; y < GGO.board_width; y++) {
    for (x = 0; x < GGO.board_width; x++) {

      int stone = BOARD(x, y);
      if (stone == EMPTY) continue;

      int stone_is_black = (stone == BLACK);
      int last_move = ((GGO.last_move_x == x) && (GGO.last_move_y == y));

      psp_go_get_pix_xy(x, y, &x_pix, &y_pix);
      psp_sdl_blit_node_xy(x_pix, y_pix, stone_is_black, last_move, board_size_id);
    }
  }

  if (! psp_go_is_computer_move()) {
    psp_go_get_pix_xy(GGO.current_x, GGO.current_y, &x_pix, &y_pix);
    psp_sdl_blit_ghost_xy(x_pix, y_pix, GGO.gameinfo.to_move == BLACK, board_size_id);
  }
}

char *
psp_go_get_move_str()
{
  SGFNode *last_node; 
  int      valid_xy;
  int      letter;
  int      x = 0;
  int      y = 0;

  static char buffer[32];
  strcpy(buffer, "NC");

  last_node = GGO.sgftree.lastnode;
  if (last_node) {
    valid_xy = gnugo_get_sgfnode_xy(last_node, &x, &y);
    if (! valid_xy) strcpy(buffer, "pass");
    else {
      letter = 'A' + x;
      if (letter >= 'I') letter++;
      sprintf(buffer, "%c%d", letter, y+1);
    }
  }
  return buffer;
}

char *
psp_go_get_comment_str()
{
  SGFNode *last_node = GGO.sgftree.lastnode;
  if (last_node) {
    return gnugo_get_sgfnode_comment(last_node);
  }
  return NULL;
}

void
psp_go_display_status()
{
  char  buffer[128];

  int b_x    =   5;
  int width  = 120;
  int height = 200;

  int y_orig =  5;
  int max_y  = y_orig + height - 10;
  int b_color = psp_sdl_rgb(0x3a, 0x3a, 0);

  psp_sdl_draw_rectangle( b_x    ,  y_orig    , width    , height    , 0, 0);
  psp_sdl_back3_rectangle( b_x + 1,  y_orig + 1, width - 2, height - 2);

  b_x += 5;
  y_orig += 10;

  /* Black/White to play */
  if (GGO.go_game_mode == GGO_HUMAN_VS_COMPUTER) strcpy(buffer, "Human vs PSP");
  else
  if (GGO.go_game_mode == GGO_HUMAN_VS_HUMAN)    strcpy(buffer, "Human vs Human");
  else                                           strcpy(buffer, "PSP vs PSP");

  psp_sdl_print(b_x, y_orig, buffer, GP2X_MENU_WARNING_COLOR);
  y_orig += 20;

  if (GGO.gameinfo.to_move == BLACK) {
    psp_sdl_print(b_x, y_orig, "Black to play", GP2X_MENU_BLACK_COLOR);
  } else {
    psp_sdl_print(b_x, y_orig, "White to play", GP2X_MENU_WHITE_COLOR);
  }
  y_orig += 20;

  if (GGO.game_over) {
    psp_sdl_print(b_x, y_orig, "GAME OVER !", GP2X_MENU_WARNING_COLOR);
  } else 
  if (psp_go_is_computer_move()) {
    psp_sdl_print(b_x, y_orig, "PSP Thinking", GP2X_MENU_RED_COLOR);
  }
  y_orig += 20;
  
  /* Move XXX  A1 / Pass etc ... */
  psp_sdl_print(b_x, y_orig, "Last move :", GP2X_MENU_WHITE_COLOR);
  y_orig += 10;
  sprintf(buffer, "[%d] %s", movenum, psp_go_get_move_str());
  psp_sdl_print(b_x, y_orig, buffer, GP2X_MENU_TEXT_COLOR);
  y_orig += 20;

  psp_sdl_print(b_x, y_orig, "White :", GP2X_MENU_WHITE_COLOR);
  y_orig += 10;
  sprintf(buffer, "%d captured", white_captured);
  psp_sdl_print(b_x, y_orig, buffer, GP2X_MENU_TEXT_COLOR);
  y_orig += 20;

  psp_sdl_print(b_x, y_orig, "Black :", GP2X_MENU_BLACK_COLOR);
  y_orig += 10;
  sprintf(buffer, "%d captured", black_captured);
  psp_sdl_print(b_x, y_orig, buffer, GP2X_MENU_TEXT_COLOR);
  y_orig += 20;


  char *comment = psp_go_get_comment_str();
  char *scan = NULL;
  int length = 0;

  if (comment) {
    while (y_orig < max_y - 12) {
      length = strlen(comment);
      if (length > 15) {
        strncpy(buffer, comment, 15);
        buffer[15] = 0;
        scan = strrchr(buffer, ' ');
        if (scan) {
          *scan = 0;
          comment += (scan - buffer) + 1;
        } else {
          comment += 15;
        }
        psp_sdl_print(b_x, y_orig, buffer, GP2X_MENU_TEXT_COLOR);
      } else {
        strcpy(buffer, comment);
        psp_sdl_print(b_x, y_orig, buffer, GP2X_MENU_TEXT_COLOR);
        break;
      }
      y_orig += 10;
    }
  }

# if 0 //LUDO: TOO LONG !
  psp_sdl_print(x_orig, y_orig, "Estimated score :", GP2X_MENU_BLACK_COLOR);
  y_orig += 10;
  gnugo_estimate_score(&upper_bound, &lower_bound);
  score = (int) ((lower_bound + upper_bound) / 2.0);
  if (fabs(score) < 1e-3) sprintf(buffer, "Even !");
  if (score       >  0.0) sprintf(buffer, "White %d", (int)score);
  else                    sprintf(buffer, "Black %d", (int)-score);
  psp_sdl_print(x_orig, y_orig, buffer, GP2X_MENU_TEXT_COLOR);
  y_orig += 20;
# endif
}

void
psp_go_refresh_screen()
{
  psp_go_display_board();
  if (psp_go_status_mode) {
    psp_go_display_status();
  }
}

void
psp_go_go_down()
{
  GGO.current_y--; 
  if (GGO.current_y < 0) GGO.current_y = GGO.board_width - 1;
  psp_go_board_dirty = 1;
}

void
psp_go_go_up()
{
  GGO.current_y++; 
  if (GGO.current_y >= GGO.board_width) GGO.current_y = 0;
  psp_go_board_dirty = 1;
}

void
psp_go_go_left()
{
  GGO.current_x--; 
  if (GGO.current_x < 0) GGO.current_x = GGO.board_width - 1;
  psp_go_board_dirty = 1;
}

void
psp_go_go_right()
{
  GGO.current_x++; 
  if (GGO.current_x >= GGO.board_width) GGO.current_x = 0;
  psp_go_board_dirty = 1;
}

void
psp_go_new_game()
{
  psp_global_reset();
  psp_play_reset();
}


void
psp_go_game_over()
{
  psp_go_refresh_screen();

  GGO.game_over = 1;

  float score = gnugo_estimate_score(NULL, NULL);
  char  buffer[128];

  if (score == 0.0) sprintf(buffer, "No winner, Jigo");
  if (score >  0.0) sprintf(buffer, "White wins with %d points", (int)score);
  else              sprintf(buffer, "Black wins with %d points", (int)-score);

  psp_sdl_back3_rectangle(15, 80, 280, 45);
  psp_sdl_draw_rectangle(15,  80, 280, 45, 0, 0);
  psp_sdl_print(25,   95, buffer, GP2X_MENU_RED_COLOR);
  psp_sdl_print(25,  105, "Game over press X to continue ...", GP2X_MENU_RED_COLOR);
  psp_sdl_flip();

  psp_kbd_wait_no_button();
  while (1)
  {
    gp2xCtrlData c;
    gp2xCtrlReadBufferPositive(&c, 1);
    if (c.Buttons & GP2X_CTRL_CROSS) break;
  }
  psp_kbd_wait_no_button();

  psp_go_board_dirty = 1;
  psp_go_set_move_mode(0);

  GGO.passes = 0;
}

void
psp_go_forward_move()
{
  if (sgftreeForward(&GGO.sgftree)) {
    GGO.gameinfo.to_move = gnugo_play_sgfnode(GGO.sgftree.lastnode, GGO.gameinfo.to_move);
    psp_go_board_dirty = 1;
  }
}

void
psp_go_undo_move()
{
  if (gnugo_undo_move(1)) {
    sgftreeBack(&GGO.sgftree);
    GGO.gameinfo.to_move = OTHER_COLOR(GGO.gameinfo.to_move);
    psp_go_board_dirty = 1;
  }
}

void
psp_go_first_move()
{
  gameinfo_play_sgftree(&GGO.gameinfo, &GGO.sgftree, "1");
  psp_go_board_dirty = 1;
}

void
psp_go_last_move()
{
  while (sgftreeForward(&GGO.sgftree)) {
    GGO.gameinfo.to_move = gnugo_play_sgfnode(GGO.sgftree.lastnode, GGO.gameinfo.to_move);
  }
  psp_go_board_dirty = 1;
}

void
psp_go_pass()
{
  if (GGO.game_over) return;

  if (GGO.go_game_mode != GGO_COMPUTER_VS_COMPUTER) {
    if (psp_play_player_pass(0)) {
      psp_go_board_dirty = 1;
    }
  }
}

void
psp_go_move()
{
  if (GGO.game_over) return;

  if (GGO.go_game_mode != GGO_COMPUTER_VS_COMPUTER) {
    if (psp_play_player_move()) {
      psp_go_board_dirty = 1;
    }
  }
}

void
psp_go_resign_game()
{
  if (GGO.game_over) return;

  GGO.passes = 2;
}

void
psp_go_main_loop()
{
  gp2xCtrlData c;
  long        new_pad;
  long        old_pad;
  int         last_time;

  old_pad   = 0;
  last_time = 0;

  psp_main_menu();

  while (1) {

    do {

      if (GGO.psp_screenshot_mode) {
        GGO.psp_screenshot_mode = 0;
        psp_go_refresh_screen();
        psp_sdl_flip();
        psp_go_refresh_screen();
        psp_sdl_flip();
        psp_sdl_save_screenshot();
      }

      if (psp_go_board_dirty) {
        psp_go_board_dirty = 0;
        psp_go_refresh_screen();
        psp_sdl_flip();
# if defined(DINGUX_MODE)
        // avoid issue with cache 
        psp_go_refresh_screen();
        psp_sdl_flip();
        psp_go_refresh_screen();
        psp_sdl_flip();
        psp_go_refresh_screen();
        psp_sdl_flip();
# endif
      }

      if (psp_go_is_computer_move()) {

        if (! psp_play_computer_move()) {
          /* Game OVER ! */
          GGO.passes = 2;
        } else {
          psp_go_refresh_screen();
          psp_sdl_flip();
        }
        if (GGO.go_game_mode == GGO_COMPUTER_VS_COMPUTER) {
          GGO.gameinfo.computer_player = OTHER_COLOR(GGO.gameinfo.computer_player);
        }
      }

      if (GGO.passes >= 2) {
        /* Game OVER ! */
        psp_go_game_over();
      }

    } while (psp_go_board_dirty);

    while (1)
    {
      gp2xCtrlPeekBufferPositive(&c, 1);

      if (c.Buttons)  break;

      if (GGO.go_game_mode == GGO_COMPUTER_VS_COMPUTER) {
        break;
      }
    }

    new_pad = c.Buttons;

    if (new_pad) {
      if ((old_pad != new_pad) || ((c.TimeStamp - last_time) > PSP_GO_MIN_TIME)) {
        last_time = c.TimeStamp;
        old_pad = new_pad;
      } else continue;
    }

    if ((c.Buttons & (GP2X_CTRL_LTRIGGER|GP2X_CTRL_RTRIGGER|GP2X_CTRL_START)) ==
        (GP2X_CTRL_LTRIGGER|GP2X_CTRL_RTRIGGER|GP2X_CTRL_START)) {
      /* Exit ! */
      psp_sdl_exit(0);
    } else
    if((new_pad & GP2X_CTRL_START) == GP2X_CTRL_START) {
      psp_go_status_mode = ! psp_go_status_mode;
      psp_go_board_dirty = 1;
      psp_kbd_wait_no_button();

    } else
    if((new_pad & GP2X_CTRL_SELECT) == GP2X_CTRL_SELECT) {
      psp_main_menu();
      psp_go_board_dirty = 1;
    } else {

      if (new_pad & GP2X_CTRL_CIRCLE) {
        psp_go_move_mode = 1;
        psp_go_move();
        psp_kbd_wait_no_button();
      } else
      if (new_pad & GP2X_CTRL_CROSS) {
        psp_go_move_mode = 1;
        psp_go_pass();
        psp_kbd_wait_no_button();
      } else
      if ((new_pad & GP2X_CTRL_LTRIGGER) == GP2X_CTRL_LTRIGGER) {
        psp_go_move_mode = 0;
        psp_go_first_move();
        psp_kbd_wait_no_button();
      } else
      if (new_pad & GP2X_CTRL_SQUARE) {
        psp_go_move_mode = 0;
        psp_go_undo_move();
        psp_kbd_wait_no_button();
      } else
      if ((new_pad & GP2X_CTRL_RTRIGGER) == GP2X_CTRL_RTRIGGER) {
        psp_go_move_mode = 0;
        psp_go_last_move();
        psp_kbd_wait_no_button();
      } else
      if (new_pad & GP2X_CTRL_TRIANGLE) {
        psp_go_move_mode = 0;
        psp_go_forward_move();
        psp_kbd_wait_no_button();
      }
      if (new_pad & GP2X_CTRL_UP) {
        psp_go_go_up();
      }
      if (new_pad & GP2X_CTRL_DOWN) {
        psp_go_go_down();
      }
      if (new_pad & GP2X_CTRL_LEFT) {
        psp_go_go_left();
      }
      if (new_pad & GP2X_CTRL_RIGHT) {
        psp_go_go_right();
      }
    }
  }
}

int
psp_go_load_sgf(char *filename)
{
  psp_global_reset();
  psp_play_init_all(1);

	if (!sgftree_readfile(&GGO.sgftree, filename)) {
    return 1;
	}

  /* to avoid changing handicap etc. */
  GGO.sgf_initialized = 1;
  gameinfo_load_sgfheader(&GGO.gameinfo, GGO.sgftree.root);
  gameinfo_play_sgftree(&GGO.gameinfo, &GGO.sgftree, NULL);
	sgfOverwritePropertyInt(GGO.sgftree.root, "HA", GGO.gameinfo.handicap);

  GGO.board_width = board_size;
  GGO.go_handicap = GGO.gameinfo.handicap;

  return 0;
}

int
psp_go_save_sgf(char *filename)
{
  int ok = 0;
  /* make sure we are saving proper handicap */
  psp_play_init_sgf();
  ok = writesgf(GGO.sgftree.root, filename);
  return ! ok;
}

int 
SDL_main(int argc,char *argv[])
{
  psp_sdl_init();
  psp_global_init();
  psp_play_init();

  psp_go_main_loop();

  psp_sdl_exit(0);

  return 0;
}

