/*
 *  Copyright (C) 2009 Ludovic Jacomme (ludovic.jacomme@gmail.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <SDL/SDL.h>

#include "gp2x_psp.h"
#include "gp2x_cpu.h"

#include "psp_global.h"
#include "psp_danzeff.h"
#include "psp_sdl.h"
#include "psp_menu.h"
#include "psp_fmgr.h"
#include "psp_menu_help.h"
#include "psp_kbd.h"
#include "psp_go.h"

extern SDL_Surface *back_surface;

# define MENU_FILENAME     0
# define MENU_GAME_MODE    1
# define MENU_BOARD_SIZE   2
# define MENU_PLAYER_COLOR 3
# define MENU_HANDICAP     4
# define MENU_KOMI         5
# define MENU_LEVEL        6
# define MENU_CLOCK        7
# define MENU_SCREENSHOT   8

# define MENU_HELP         9

# define MENU_NEW_GAME    10
# define MENU_RESIGN_GAME 11
# define MENU_LOAD_SGF    12
# define MENU_SAVE_SGF    13
# define MENU_BACK        14
# define MENU_SAVE_CFG    15
# define MENU_EXIT        16

# define MAX_MENU_ITEM (MENU_EXIT + 1)

  static menu_item_t menu_list[] =
  { 
    { "Filename        : " },
    { "Game mode       : " },
    { "Board size      : " },
    { "Player color    : " },
    { "Handicap        : " },
    { "Komi            : " },
    { "Level           : " },
    { "Clock frequency : " },
    { "Save Screenshot : " },

    { "Help" },

    { "New game" },
    { "Resign game" },
    { "Load SGF" },
    { "Save SGF" },
    { "Back to game" },
    { "Save config" },
    { "Exit" },
  };

  static int cur_menu_id = MENU_BACK;

  static int go_board_size   = 19;
  static int go_player_color = GGO_BLACK;
  static int go_handicap     = 0;
  static int go_level        = 10;
  static int go_komi         = 0;
  static int go_game_mode    = GGO_HUMAN_VS_COMPUTER;
  static int psp_cpu_clock   = GP2X_DEF_CLOCK;
# define FILENAME_FIELD_WIDTH      25

void
string_fill_with_space(char *buffer, int size)
{
  int length = strlen(buffer);
  int index;

  for (index = length; index < size; index++) {
    buffer[index] = ' ';
  }
  buffer[size] = 0;
}

static void 
psp_display_screen_menu(void)
{
  char buffer[128];
  int menu_id = 0;
  int color   = 0;
  int x       = 0;
  int y       = 0;
  int y_step  = 0;

  psp_sdl_blit_background();
  
  x      = 10;
  y      = 15;
  y_step = 10;

  for (menu_id = 0; menu_id < MAX_MENU_ITEM; menu_id++) {
    color = GP2X_MENU_TEXT_COLOR;
    if (cur_menu_id == menu_id) color = GP2X_MENU_SEL_COLOR;
    else 
    if (menu_id == MENU_EXIT) color = GP2X_MENU_RED_COLOR;
    else
    if (menu_id == MENU_HELP) color = GP2X_MENU_GREEN_COLOR;

    psp_sdl_back2_print(x, y, menu_list[menu_id].title, color);

    if (menu_id == MENU_FILENAME) {
      int length = strlen(GGO.game_filename);
      int first = 0;
      if (length > (FILENAME_FIELD_WIDTH-1)) {
        first = length - (FILENAME_FIELD_WIDTH-1);
      }
      sprintf(buffer, "%s", GGO.game_filename + first);
      if (menu_id == cur_menu_id) strcat(buffer, "_");
      string_fill_with_space(buffer, FILENAME_FIELD_WIDTH+1);
      psp_sdl_back2_print(130, y, buffer, color);

    } else
    if (menu_id == MENU_BOARD_SIZE) {
      sprintf(buffer,"%d", go_board_size);
      string_fill_with_space(buffer, 4);
      psp_sdl_back2_print(130, y, buffer, color);
    } else
    if (menu_id == MENU_PLAYER_COLOR) {
      if (go_player_color == GGO_BLACK) strcpy(buffer, "black");
      else                              strcpy(buffer, "white");
      string_fill_with_space(buffer, 5);
      psp_sdl_back2_print(130, y, buffer, color);
    } else
    if (menu_id == MENU_HANDICAP) {
      sprintf(buffer,"%d", go_handicap);
      string_fill_with_space(buffer, 4);
      psp_sdl_back2_print(130, y, buffer, color);
    } else
    if (menu_id == MENU_KOMI) {
      sprintf(buffer,"%.1f", ((float)go_komi) / 10.0);
      string_fill_with_space(buffer, 6);
      psp_sdl_back2_print(130, y, buffer, color);
    } else
    if (menu_id == MENU_GAME_MODE) {
      if (go_game_mode == GGO_HUMAN_VS_COMPUTER) strcpy(buffer, "Human vs CPU");
      else
      if (go_game_mode == GGO_HUMAN_VS_HUMAN) strcpy(buffer, "Human vs Human");
      else                                    strcpy(buffer, "CPU vs CPU");
      string_fill_with_space(buffer, 15);
      psp_sdl_back2_print(130, y, buffer, color);
    } else
    if (menu_id == MENU_LEVEL) {
      sprintf(buffer,"%d", go_level);
      string_fill_with_space(buffer, 4);
      psp_sdl_back2_print(130, y, buffer, color);
    } else
    if (menu_id == MENU_CLOCK) {
      sprintf(buffer,"%d", psp_cpu_clock);
      string_fill_with_space(buffer, 4);
      psp_sdl_back2_print(130, y, buffer, color);
    } else
    if (menu_id == MENU_SCREENSHOT) {
      sprintf(buffer,"%d", GGO.psp_screenshot_id);
      string_fill_with_space(buffer, 4);
      psp_sdl_back2_print(130, y, buffer, color);
      y += y_step;
    } else
    if (menu_id == MENU_SAVE_CFG) {
      y += y_step;
    } else
    if (menu_id == MENU_HELP) {
      y += y_step;
    } else
    if (menu_id == MENU_EXIT) {
      y += y_step;
    }

    y += y_step;
  }
}

void
psp_main_menu_new_game(void)
{
  psp_go_new_game();
}

void
psp_main_menu_resign_game(void)
{
  psp_go_resign_game();
}

int
psp_main_menu_confirm_overwrite()
{
  gp2xCtrlData c;

  psp_display_screen_menu();
  psp_sdl_back2_print(140, 160, "B overwrite X cancel", GP2X_MENU_WARNING_COLOR);
  psp_sdl_flip();
  psp_kbd_wait_no_button();

  do
  {
    gp2xCtrlReadBufferPositive(&c, 1);
    c.Buttons &= GP2X_ALL_BUTTON_MASK;

    if (c.Buttons & GP2X_CTRL_CROSS) {
      return 1;
    } else 
    if (c.Buttons & GP2X_CTRL_TRIANGLE) {
      return 0;
    }

  } while (c.Buttons == 0);
  psp_kbd_wait_no_button();

  return 0;
}

void
psp_main_menu_load_sgf(void)
{
  int ret;

  ret = psp_fmgr_menu_load(GP2X_FMGR_FORMAT_SGF);
  if (ret == 1) /* load OK */
  {
	  psp_display_screen_menu();
    psp_sdl_back2_print(140, 160, "File loaded !", 
                       GP2X_MENU_NOTE_COLOR);
	  psp_sdl_flip();
	  sleep(1);
	
	  psp_go_set_move_mode(0);
  }
  else 
  if (ret == -1) /* Load Error */
  {
	  psp_display_screen_menu();
    psp_sdl_back2_print(140, 160, "Can't load file !", GP2X_MENU_WARNING_COLOR);
	  psp_sdl_flip();
	  sleep(1);
  }

  go_board_size = GGO.board_width;
  go_handicap   = GGO.go_handicap;
}

void
psp_main_menu_filename_add_char(int a_char)
{
  int length = strlen(GGO.game_filename);
  if (length < (MAX_PATH-2)) {
    GGO.game_filename[length] = a_char;
    GGO.game_filename[length+1] = 0;
  }
}

void
psp_main_menu_filename_clear()
{
  GGO.game_filename[0] = 0;
}

void
psp_main_menu_filename_del()
{
  int length = strlen(GGO.game_filename);
  if (length >= 1) {
    GGO.game_filename[length - 1] = 0;
  }
}



static int
psp_main_menu_save_sgf()
{
  int error;
  char sgf_filename[MAX_PATH];

  if (GGO.game_filename[0] == 0) {
    strcpy(GGO.game_filename, "newgame");
  }

  char* scan_char = strrchr(GGO.game_filename, '.');
  if (! scan_char) {
    strcat(GGO.game_filename, ".sgf");
  }
  strcpy(sgf_filename, "./sgf/");
  strcat(sgf_filename, GGO.game_filename);

  int do_write = 0;

  struct stat aStat;
  if (stat(sgf_filename, &aStat)) {
    do_write = 1;
  } else {
    do_write = psp_main_menu_confirm_overwrite();
  }
  if (! do_write) {
    return 0;
  }
  error = psp_go_save_sgf(sgf_filename);

  if (! error) /* save OK */
  {
    psp_display_screen_menu();
    psp_sdl_back2_print(140, 160, "File saved !", 
                       GP2X_MENU_NOTE_COLOR);
    psp_sdl_flip();
    sleep(1);
  }
  else 
  {
    psp_display_screen_menu();
    psp_sdl_back2_print(140, 160, "Can't save file !", 
                       GP2X_MENU_WARNING_COLOR);
    psp_sdl_flip();
    sleep(1);
  }
  return error;
}


static void
psp_main_menu_validate(void)
{
  int need_to_reset = 0;
  GGO.psp_cpu_clock = psp_cpu_clock;
 
  if (go_handicap != GGO.go_handicap) {
    GGO.go_handicap = go_handicap;
  }
  if (go_komi != GGO.go_komi) {
    GGO.go_komi = go_komi;
  }
  if (go_game_mode != GGO.go_game_mode) {
    GGO.go_game_mode = go_game_mode;
  }
  if (go_level != GGO.go_level) {
    GGO.go_level = go_level;
  }
  if (go_player_color != GGO.go_player_color) {
    GGO.go_player_color = go_player_color;
    need_to_reset = 1;
  }
  if (go_board_size != GGO.board_width) {
    GGO.board_width = go_board_size;
    need_to_reset = 1;
  }

  if (need_to_reset) {
    psp_main_menu_new_game();
  }
  gp2xPowerSetClockFrequency(GGO.psp_cpu_clock);
}


static void
psp_main_menu_save_config()
{
  int error;

  psp_main_menu_validate();
  error = psp_save_configuration();

  if (! error) /* save OK */
  {
    psp_display_screen_menu();
    psp_sdl_back2_print(140, 160, "File saved !", 
                       GP2X_MENU_NOTE_COLOR);
    psp_sdl_flip();
    sleep(1);
  }
  else 
  {
    psp_display_screen_menu();
    psp_sdl_back2_print(140, 160, "Can't save file !", 
                       GP2X_MENU_WARNING_COLOR);
    psp_sdl_flip();
    sleep(1);
  }
}

static void
psp_main_menu_screenshot(void)
{
  GGO.psp_screenshot_mode = 1;
}

int
psp_main_menu_exit(void)
{
  gp2xCtrlData c;

  psp_display_screen_menu();
  psp_sdl_back2_print(140, 160, "press B to confirm !", GP2X_MENU_WARNING_COLOR);
  psp_sdl_flip();

  psp_kbd_wait_no_button();

  do
  {
    gp2xCtrlReadBufferPositive(&c, 1);
    c.Buttons &= GP2X_ALL_BUTTON_MASK;

    if (c.Buttons & GP2X_CTRL_CROSS) psp_sdl_exit(0);

  } while (c.Buttons == 0);

  psp_kbd_wait_no_button();

  return 0;
}

static void
psp_main_menu_init(void)
{
  psp_cpu_clock    = GGO.psp_cpu_clock;
  go_handicap      = GGO.go_handicap;
  go_komi          = GGO.go_komi;
  go_game_mode     = GGO.go_game_mode;
  go_level         = GGO.go_level;
  go_player_color  = GGO.go_player_color;
  go_board_size    = GGO.board_width;
}

void
psp_main_menu_board_size( int step )
{
  if (step > 0) {
    if (go_board_size < 19) go_board_size++;
    else                    go_board_size = 5;
  } else {
    if (go_board_size > 5) go_board_size--;
    else                   go_board_size = 19;
  }
}

void
psp_main_menu_player_color( int step )
{
    if (go_player_color == GGO_BLACK) go_player_color = GGO_WHITE;
    else                              go_player_color = GGO_BLACK;
}

void
psp_main_menu_game_mode( int step )
{
  if (step > 0) {
    if (go_game_mode < GGO_COMPUTER_VS_COMPUTER) go_game_mode++;
    else                                         go_game_mode = 0;
  } else {
    if (go_game_mode > 0) go_game_mode--;
    else                  go_game_mode = GGO_COMPUTER_VS_COMPUTER;
  }
}

void
psp_main_menu_handicap( int step )
{
  if (step > 0) {
    if (go_handicap < 9) go_handicap++;
    else                 go_handicap = 0;
  } else {
    if (go_handicap > 0) go_handicap--;
    else                 go_handicap = 9;
  }
}

void
psp_main_menu_komi( int step )
{
  if (step > 0) {
    if (go_komi < 100) go_komi += 5;
    else               go_komi  = 0;
  } else {
    if (go_komi >  0)  go_komi -= 5;
    else               go_komi  = 100;
  }
}

void
psp_main_menu_level( int step )
{
  if (step > 0) {
    if (go_level < 10) go_level++;
    else               go_level = 0;
  } else {
    if (go_level > 0) go_level--;
    else              go_level = 10;
  }
}

void
psp_main_menu_clock( int step )
{
  if (step > 0) {
    psp_cpu_clock += 10;
    if (psp_cpu_clock > GP2X_MAX_CLOCK) psp_cpu_clock = GP2X_MAX_CLOCK;
  } else {
    psp_cpu_clock -= 10;
    if (psp_cpu_clock < GP2X_MIN_CLOCK) psp_cpu_clock = GP2X_MIN_CLOCK;
  }
}

int 
psp_main_menu(void)
{
  gp2xCtrlData c;
  long        new_pad;
  long        old_pad;
  int         last_time;
  int         end_menu;

  int         danzeff_mode = 0;
  int         danzeff_key;

  int         step = 0;

  psp_kbd_wait_no_button();

  old_pad   = 0;
  last_time = 0;
  end_menu  = 0;

  psp_main_menu_init();

  while (! end_menu)
  {
    psp_display_screen_menu();

    if (danzeff_mode) {
      danzeff_moveTo(-20, -30);
      danzeff_render();
    }
    psp_sdl_flip();

    gp2xCtrlPeekBufferPositive(&c, 1);
    c.Buttons &= GP2X_ALL_BUTTON_MASK;

    new_pad = c.Buttons;

    if ((old_pad != new_pad) || ((c.TimeStamp - last_time) > GP2X_MENU_MIN_TIME)) {
      last_time = c.TimeStamp;
      old_pad = new_pad;

    } else continue;

    if (danzeff_mode) {

      danzeff_key = danzeff_readInput(c);
      if (danzeff_key > DANZEFF_START) {
        /* Disable utf8 */
        danzeff_key &= 0x7f;
        if (danzeff_key > ' ') {
          psp_main_menu_filename_add_char(danzeff_key);
        } else
        if (danzeff_key == DANZEFF_CLEAR) {
          psp_main_menu_filename_clear();
        } else
        if (danzeff_key == DANZEFF_DEL) {
          psp_main_menu_filename_del();
        }

      } else 
      if ((danzeff_key == DANZEFF_START ) || 
          (danzeff_key == DANZEFF_SELECT)) 
      {
        danzeff_mode = 0;
        old_pad = new_pad = 0;

        psp_kbd_wait_no_button();
      }
      continue;
    }

    if ((c.Buttons & (GP2X_CTRL_LTRIGGER|GP2X_CTRL_RTRIGGER|GP2X_CTRL_START)) ==
        (GP2X_CTRL_LTRIGGER|GP2X_CTRL_RTRIGGER|GP2X_CTRL_START)) {
      /* Exit ! */
      psp_sdl_exit(0);
    } else
    if ((new_pad == GP2X_CTRL_LEFT ) || 
        (new_pad == GP2X_CTRL_RIGHT) ||
        (new_pad == GP2X_CTRL_CROSS) || 
        (new_pad == GP2X_CTRL_CIRCLE))
    {
      step = 0;

      if (new_pad & GP2X_CTRL_RIGHT) {
        step = 1;
      } else
      if (new_pad & GP2X_CTRL_LEFT) {
        step = -1;
      }

      switch (cur_menu_id ) 
      {
        case MENU_BOARD_SIZE   : psp_main_menu_board_size( step );
        break;
        case MENU_PLAYER_COLOR : psp_main_menu_player_color( step );
        break;
        case MENU_HANDICAP     : psp_main_menu_handicap( step );
        break;
        case MENU_LEVEL        : psp_main_menu_level( step );
        break;
        case MENU_KOMI         : psp_main_menu_komi( step );
        break;
        case MENU_GAME_MODE    : psp_main_menu_game_mode( step );
        break;
        case MENU_CLOCK        : psp_main_menu_clock( step );
        break;
        case MENU_SCREENSHOT   : psp_main_menu_screenshot();
                                 end_menu = 1;
        break;              
        case MENU_NEW_GAME     : psp_main_menu_new_game();
                                 end_menu = 1;
        break;                 
        case MENU_RESIGN_GAME  : psp_main_menu_resign_game();
                                 end_menu = 1;
        break;                 
        case MENU_LOAD_SGF     : psp_main_menu_load_sgf();
        break;                 
        case MENU_SAVE_SGF     : psp_main_menu_save_sgf();
        break;                 
        case MENU_SAVE_CFG     : psp_main_menu_save_config();
        break;               
        case MENU_BACK         : end_menu = 1;
        break;                 
        case MENU_EXIT         : psp_main_menu_exit();
        break;                 
        case MENU_HELP         : psp_help_menu();
                                 old_pad = new_pad = 0;
        break;              
        case MENU_FILENAME     : if (step < 0) psp_main_menu_filename_del();
        break;
      }

    } else
    if(new_pad & GP2X_CTRL_START) {
      danzeff_mode = 1;
    } else
    if(new_pad & GP2X_CTRL_UP) {

      if (cur_menu_id > 0) cur_menu_id--;
      else                 cur_menu_id = MAX_MENU_ITEM-1;

    } else
    if(new_pad & GP2X_CTRL_DOWN) {

      if (cur_menu_id < (MAX_MENU_ITEM-1)) cur_menu_id++;
      else                                 cur_menu_id = 0;

    } else  
    if(new_pad & GP2X_CTRL_SQUARE) {
      /* Cancel */
      end_menu = -1;
    } else 
    if((new_pad & GP2X_CTRL_SELECT) == GP2X_CTRL_SELECT) {
      /* Back to game */
      end_menu = 1;
    }
  }

  if (end_menu > 0) {
    psp_main_menu_validate();
  }
 
  psp_kbd_wait_no_button();

  psp_sdl_clear_screen( GP2X_MENU_BLACK_COLOR );
  psp_sdl_flip();
  psp_sdl_clear_screen( GP2X_MENU_BLACK_COLOR );
  psp_sdl_flip();

  return 1;
}

